from Products.CMFCore import FSPageTemplate as cmfFSPT
from Products.CMFCore.FSPageTemplate import *
from ZopeAltPageTemplate import ZopePageTemplate, compiler_registry
from Products.CMFCore import utils
from OFS.SimpleItem import SimpleItem
import Globals, AccessControl
from ZODB.PersistentMapping import PersistentMapping
from Acquisition import Explicit

FSPT = FSPageTemplate

class MacroCollection(PersistentMapping, Explicit):
    __altpt_0_2__ = 1
    
    def __getitem__(self, name):
        self.aq_parent._updateFromFS()
        try:
            return PersistentMapping.__getitem__(self, name)
        except KeyError:
            raise NameError, name

class FSPageTemplate(ZopePageTemplate, FSObject, Script):
    "Wrapper for Page Template"
     
    meta_type = 'Alternative Filesystem Page Template'
    _icon = 'misc_/AltPT/fsaltpt.png'
    icon = _icon
    _compiled = None
    # FIXME: is there already a convention for this?
    encoding = 'latin1'

    manage_options=(
        (
            {'label':'Customize', 'action':'manage_main'},
            {'label':'Test', 'action':''},
            )
        )

    security = ClassSecurityInfo()
    security.declareObjectProtected(View)

    security.declareProtected(ViewManagementScreens, 'manage_main')
    manage_main = FSPT.manage_main

    # Declare security for unprotected PageTemplate methods.
    security.declarePrivate('pt_edit', 'write', 'pt_editForm',
                            'pt_editAction', 'pt_setTitle', 'pt_edit',
                            'pt_expand', 'pt_upload',
                            'pt_changePrefs', 'PUT', 'manage_FTPput',
                            'write', 'manage_historyCopy',
                            'manage_beforeHistoryCopy',
                            'manage_afterHistoryCopy')

    def __init__(self, id, filepath, fullname=None, properties=None):
        FSObject.__init__(self, id, filepath, fullname, properties)
        self.ZBindings_edit(self._default_bindings)
        self.macros = MacroCollection()
        self.macros.__roles__ = None
        self.macros.__allow_access_to_unprotected_subobjects__ = 1

    def manage_afterAdd(self, item, container):
        # do nothing here for FSObjects
        pass

    def _createZODBClone(self):
        """Create a ZODB (editable) equivalent of this object."""
        t = ZopePageTemplate(self.getId(), self.read())
        t.handlers = self.handlers
        return t

    def ZCacheable_isCachingEnabled(self):
        return 0

    def _readFile(self, reparse):
        fp = expandpath(self._filepath)
        file = open(fp, 'rb')
        try: data = file.read()
        finally: file.close()
        if reparse:
            self._compiled = repr(compiler_registry)
            # parse #: commands
            # WARNING: you can seriously break zope with these
            if data.startswith('#:'):
                d = data.split('\n')
                while d[0].startswith('#:'):
                    c, arg = d[0].split(None, 1)
                    del d[0]
                    if arg.startswith('='):
                        # property/attribute set
                        arg = eval(arg[1:])
                        setattr(self, c, arg)
                    # otherwise "c" is a command
                    # no commands are defined yet
                data = '\n'.join(d)
            self.write(data)

    security.declarePrivate('read')
    def read(self):
        # Tie in on an opportunity to auto-update
        self._updateFromFS()
        return ZopePageTemplate.read(self)

    def paxtree(self):
        'fetch the pax tree from the zodb'
        if self._compiled != repr(compiler_registry):
            self._parsed = None
        self._updateFromFS()
        return ZopePageTemplate.paxtree(self)


Globals.InitializeClass(FSPageTemplate)
#cmfFSPT.FSPageTemplate = FSPageTemplate

registerFileExtension('pt', FSPageTemplate)
registerFileExtension('html', FSPageTemplate)
registerFileExtension('htm', FSPageTemplate)
registerMetaType('Page Template', FSPageTemplate)

def initialize(context):
    utils.registerIcon(FSPageTemplate,
                       'www/fsaltpt.png', globals())
