import AltTAL, pax, PTmetal_handler, pax.paxtransform
from Expressions import getEngine, TALESError, SecureModuleImporter
import cPickle as pickle
#import pickle

import os, AccessControl, Acquisition, sys
from Globals import DTMLFile, ImageFile, MessageDialog, package_home
from zLOG import LOG, ERROR, INFO
from OFS.SimpleItem import SimpleItem
from DateTime.DateTime import DateTime
from string import join, strip, rstrip, split, replace, lower
from Shared.DC.Scripts.Script import Script, BindingsUI
from Shared.DC.Scripts.Signature import FuncCode
from Shared.DC.Scripts.Bindings import NameAssignments, Bindings
from AccessControl import getSecurityManager
try:
    from AccessControl import Unauthorized
except ImportError:
    Unauthorized = "Unauthorized"
from OFS.History import Historical, html_diff
from OFS.Cache import Cacheable
from OFS.Traversable import Traversable
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from webdav.Lockable import ResourceLockedError
from webdav.WriteLockInterface import WriteLockInterface
from ZODB.PersistentMapping import PersistentMapping
from OFS.SimpleItem import SimpleItem
import Globals

class PTNameAssignments(NameAssignments):
    # this should emulate the following ugly zpt code:
##         c = {'template': self,
##              'here': self._getContext(),
##              'container': self._getContainer(),
##              'nothing': None,
##              'options': {},
##              'root': root,
##              'request': getattr(root, 'REQUEST', None),
##              'modules': SecureModuleImporter,
##              }
    _exprs = NameAssignments._exprs + (
              ('name_none',   'None'),
              ('name_options',   '{}'),
              ('name_root',   'self._getRoot()'),
              ('name_request',   'getattr(self, "REQUEST", None)'),
              ('name_modules',   'self._getModuleImporter()'),
              )

class PTRuntimeError(RuntimeError):
    '''The Page Template has template errors that prevent it from rendering.'''
    pass

def make_macro_collection():
    col = PersistentMapping()
    col.__roles__ = None
    col.__allow_access_to_unprotected_subobjects__ = 1
    col.__altpt_0_2__ = 1
    return col

def fetchmacros(template, tree, context):
    macros = getattr(context, 'metal_macros', {})
    col = template.macros
    col.clear()
    for name, macro in macros.items():
        col[name] = pickle.dumps(macro)
    return tree


# paxtransform handlers
handler_registry = [AltTAL.metal_handler, AltTAL.tal_handler]
# paxtransform handlers
compiler_registry = [AltTAL.metal_compiler, AltTAL.tal_compiler]
# functions!
postcompiler_registry = [fetchmacros]
# these are calculated automatically, don't touch them
handler_namespaces = {}
handler_namespaces_rev = {}

def PT_initial_namespaces():
    if len(handler_namespaces) != len(handler_registry):
        handler_namespaces.clear()
        for handler in handler_registry:
            handler_namespaces[handler.name] = handler.ns
    return handler_namespaces

source_types = {'XML':pax.xml2pax, 'HTML':pax.html2pax, 'Text':pax.text2pax}


# FIXME: lots of stuff below can probably be refactored out -
# there is still a lot of cruft pasted from ZPT

# FIXME2: it would be interesting to check if a non-zope AltPT class
# is doable. If not, all claims to its existence should be removed.

class ZopePageTemplate(Script, Historical, Cacheable,
                       Traversable, PropertyManager):
    "Zope wrapper for Alt Page Template using TAL, TALES, and METAL"
     
    __implements__ = (WriteLockInterface,)

    content_type = 'text/html'
    source_type = 'HTML'
    handlers = ('tal', 'metal')
    encoding = 'utf-8'
    _errors = ()
    _warnings = ()
    _text = ''
    _error_start = '<!-- Page Template Diagnostics'
    _paxtree = ''
    meta_type = 'Alternative Page Template'
    _icon = 'misc_/AltPT/altpt.png'

    func_defaults = None
    func_code = FuncCode((), 0)

    _default_bindings = {'name_context': 'here',
                         'name_container': 'container',
                         'name_m_self': 'template',
                         'name_subpath': 'traverse_subpath',
                         'name_none': 'nothing',
                         'name_options': 'options',
                         'name_root': 'root',
                         'name_request': 'request',
                         'name_modules': 'modules',
                         }
    _default_content_fn = os.path.join(package_home(globals()),
                                       'www', 'default.html')

    manage_options = (
        {'label':'Edit', 'action':'pt_editForm',
         'help': ('PageTemplates', 'PageTemplate_Edit.stx')},
        {'label':'Test', 'action':''},
        ) + PropertyManager.manage_options \
        + Historical.manage_options \
        + SimpleItem.manage_options \
        + Cacheable.manage_options

    _properties=({'id':'title', 'type': 'string', 'mode': 'w'},
                 {'id':'content_type', 'type':'string', 'mode': 'w'},
                 {'id':'encoding', 'type':'string', 'mode': 'w'},
                 {'id':'source_type', 'type':'selection', 'mode': 'w',
                  'select_variable': 'pt_parser_list'},
                 {'id':'handlers', 'type':'multiple selection', 'mode': 'w',
                  'select_variable': 'pt_handler_list'},
                 )

    def __init__(self, id, text=None):
        self.id = str(id)
        self.ZBindings_edit(self._default_bindings)
        if text is None:
            text = open(self._default_content_fn).read()
        self.macros = make_macro_collection()
        self.pt_edit(text)

    def manage_afterAdd(self, item, container):
        try:
            self.encoding = container.default_zope_data_encoding
        except:
            pass

    def __setstate__(self, state):
        ZopePageTemplate.inheritedAttribute('__setstate__')(self, state)
        ###backward compatibility
        col = getattr(self, 'macros', None)
        if not hasattr(col, '__altpt_0_2__'):
            self.macros = make_macro_collection()
        if not self._bind_names.isNameAssigned('here'):
            self.ZBindings_edit(self._default_bindings)
        ###/backward compatibility

    def _setupBindings(self, names={}):
        self._bind_names = names = PTNameAssignments(names)
        return names

    def _getModuleImporter(self):
        return SecureModuleImporter

    def _getRoot(self):
        root = getattr(self, 'getPhysicalRoot', None)
        # don't know how this happens, but it does: sometimes root is a string
        if callable(root):
            root = root()
        return root

    def _getContext(self):
        if hasattr(self, 'aq_parent'):
            return ZopePageTemplate.inheritedAttribute('_getContext')(self)
        return self

    def _setPropValue(self, id, value):
        PropertyManager._setPropValue(self, id, value)
        self.ZCacheable_invalidate()

    security = AccessControl.ClassSecurityInfo()

    security.declareObjectProtected('View')
    security.declareProtected('View', '__call__')

    security.declareProtected('View management screens',
      'pt_editForm', 'manage_main', 'read',
      'ZScriptHTML_tryForm', 'PrincipiaSearchSource',
      'document_src', 'pt_src')

    security.declareProtected('FTP access',
      'manage_FTPstat','manage_FTPget','manage_FTPlist')

    pt_editForm = PageTemplateFile('www/ptEdit', globals(),
                                   __name__='pt_editForm')
    pt_editForm._owner = None
    manage = manage_main = pt_editForm

    security.declareProtected('Change Page Templates',
      'pt_editAction', 'pt_setTitle', 'pt_edit', 'pt_expand',
      'pt_upload', 'pt_changePrefs')
    def pt_editAction(self, REQUEST, title, text, content_type,
                      source_type, encoding, handlers=()):
        """Change the title and document."""
        if self.wl_isLocked():
            raise ResourceLockedError, "File is locked via WebDAV"
        if self.content_type != content_type:
            self.content_type = content_type
        if self.source_type != source_type:
            self.source_type = source_type
        if self.encoding != encoding:
            self.encoding = encoding
        if self.handlers != handlers:
            self.handlers = handlers
        self.pt_setTitle(title)
        self.pt_edit(text)
        REQUEST.set('text', self.read()) # May not equal 'text'!
        message = "Saved changes."
        if getattr(self, '_warnings', None):
            message = ("<strong>Warning:</strong> <i>%s</i>" 
                       % join(self._warnings, '<br>'))
        return self.pt_editForm(manage_tabs_message=message)

    def pt_setTitle(self, title):
        self._setPropValue('title', str(title))

    def pt_upload(self, REQUEST, file=''):
        """Replace the document with the text in file."""
        if self.wl_isLocked():
            raise ResourceLockedError, "File is locked via WebDAV"
        if type(file) is not type(''): file = file.read()
        self.write(file)
        message = 'Saved changes.'
        return self.pt_editForm(manage_tabs_message=message)

    def pt_changePrefs(self, REQUEST, height=None, width=None,
                       dtpref_cols='50', dtpref_rows='20'):
        """Change editing preferences."""
        szchh = {'Taller': 1, 'Shorter': -1, None: 0}
        szchw = {'Wider': 5, 'Narrower': -5, None: 0}
        try: rows = int(height)
        except: rows = max(1, int(dtpref_rows) + szchh.get(height, 0))
        try: cols = int(width)
        except: cols = max(40, int(dtpref_cols) + szchw.get(width, 0))
        e = (DateTime('GMT') + 365).rfc822()
        setc = REQUEST['RESPONSE'].setCookie
        setc('dtpref_rows', str(rows), path='/', expires=e)
        setc('dtpref_cols', str(cols), path='/', expires=e)
        REQUEST.form.update({'dtpref_cols': cols, 'dtpref_rows': rows})
        return self.manage_main()

    def paxtree(self):
        'fetch the pax tree from the zodb'
        if not hasattr(self, '_v_paxtree'):
            self._v_paxtree = pickle.loads(self._paxtree)
        return self._v_paxtree

    def read(self, *args, **kw):
        if not self._errors:
            return self._text
                                  
        return ('%s\n %s\n-->\n%s' % (self._error_start,
                                      join(self._errors, '\n '),
                                      self._text))

    def write(self, text):
        self.ZCacheable_invalidate()
        assert type(text) in (type(''), type(u''))
        if text.startswith(self._error_start):
            errend = find(text, '-->')
            if errend >= 0:
                text = text[errend + 4:]
        if self._text != text:
            self._text = text

        # "cooking" only ever happens when the template changes
        parser = source_types[self.source_type]

        self._errors = ()
        try:
            tree = parser(text, PT_initial_namespaces(), self.encoding)
            tr_engine = pax.paxtransform.Engine()
            for handler in compiler_registry:
                tr_engine.add_handler(handler)
            # empty context:
            # compilers and postcompilers shouldn't access the context anyway
            context = getEngine().getContext({})
            context.engine = context
            context.template = self
            try:
                context.encoding = self.default_zope_data_encoding
            except:
                context.encoding = self.encoding
            tr_engine.initialize(context)
            tree = tr_engine.transform(tree, context)
            for handler in postcompiler_registry:
                tree = handler(self, tree, context)
            self._paxtree = pickle.dumps(tree)
            if hasattr(self, '_v_paxtree'):
                del self._v_paxtree
        except:
            import traceback, zLOG
            #zLOG.LOG('AltPT', 0, 'Compilation error',
            try:
                exc = sys.exc_info()
                traceback.print_exception(*exc)
                self._errors = ["Compilation failed",
                                "%s: %s" % exc[:2]]
            finally:
                del exc
                if not self._errors:
                    self._errors = ['something very wicked happened.']

    def pt_edit(self, text):
        if hasattr(text, 'read'):
            text = text.read()
        self.write(text)

    def pt_errors(self):
        return self._errors
        
    def pt_warnings(self):
        return self._warnings

    def pt_parser_list(self):
        return source_types.keys()

    def pt_handler_list(self):
        return [handler.name for handler in handler_registry]

    security.declareProtected('Change Page Templates',
      'PUT', 'manage_FTPput', 'write',
      'manage_historyCopy',
      'manage_beforeHistoryCopy', 'manage_afterHistoryCopy')

    def PUT(self, REQUEST, RESPONSE):
        """ Handle HTTP PUT requests """
        self.dav__init(REQUEST, RESPONSE)
        self.dav__simpleifhandler(REQUEST, RESPONSE, refresh=1)
        self.write(REQUEST.get('BODY', ''))
        RESPONSE.setStatus(204)
        return RESPONSE        

    manage_FTPput = PUT

    def manage_FTPget(self):
        "Get source for FTP download"
        self.REQUEST.RESPONSE.setHeader('Content-Type', self.content_type)
        return self.read()

    def get_size(self):
        return len(self.read())
    getSize = get_size

    def PrincipiaSearchSource(self):
        "Support for searching - the document's contents are searched."
        return self.read()

    def document_src(self, REQUEST=None, RESPONSE=None):
        """Return document source."""

        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')
        if REQUEST is not None and REQUEST.get('raw'):
            return self._text
        return self.read()

    def pt_src(self, REQUEST=None, RESPONSE=None):
        """Return (rendered) document source."""

        if RESPONSE is not None:
            RESPONSE.setHeader('Content-Type', 'text/plain')

        if hasattr(REQUEST, 'pax') and REQUEST.pax:
            return str(self.paxtree())
        return pax.XML(self.paxtree())

    def om_icons(self):
        """Return a list of icon URLs to be displayed by an ObjectManager"""
        icons = ({'path': self._icon,
                  'alt': self.meta_type, 'title': self.meta_type},)
        if self._errors:
            icons = icons + ({'path': 'misc_/AltPT/exclamation.gif',
                              'alt': 'Error',
                              'title': 'This template has an error'},)
        return icons

    def _exec(self, bound_names, args, kw, handlers=None, root_ns_map={}):
        """Call a Page Template"""
        __traceback_info__ = 'bindings: %s;\n args: %s;\n kw: %s' % (
            repr(bound_names), repr(args), repr(kw))
        tr_engine = pax.paxtransform.Engine()
        if handlers is None:
            handlers = [h for h in handler_registry if h.name in self.handlers]
        for handler in handlers:
            tr_engine.add_handler(handler)

        if self._errors:
            raise PTRuntimeError, 'Page Template %s has errors.' % self.id

        if not kw.has_key('args'):
            kw['args'] = args
        bound_names['options'] = kw

        security=getSecurityManager()
        bound_names['user'] = security.getUser()

        # Retrieve the value from the cache.
        keyset = None
        if self.ZCacheable_isCachingEnabled():
            # Prepare a cache key.
            keyset = {'here': self._getContext(),
                      'bound_names': bound_names}
            result = self.ZCacheable_get(keywords=keyset)
            if result is not None:
                # Got a cached value.
                return result

        # Execute the template in a new security context.
        security.addContext(self)
        try:
            #context = getEngine().getContext(self.pt_getContext(bound_names))
            context = getEngine().getContext(bound_names)
            context.engine = context
            try:
                context.encoding = self.default_zope_data_encoding
            except:
                context.encoding = self.encoding
            context.root_xmlns_map = root_ns_map
            tr_engine.initialize(context)
            result = pax.XML(tr_engine.transform(self.paxtree(), context),
                                 root_ns_map, self.encoding)
            if keyset is not None:
                # Store the result in the cache.
                self.ZCacheable_set(result, keywords=keyset)
            try:
                self.REQUEST.RESPONSE.setHeader('content-type',
                                                '%s; charset=%s' % (
                                                 self.content_type,
                                                 context.encoding))
            except AttributeError:
                pass
            return result
        finally:
            security.removeContext(self)

    def pt_expand(self, REQUEST=None):
        """Expand macros"""
        if len(handler_namespaces_rev) != len(handler_registry):
            handler_namespaces_rev.clear()
            for handler in handler_registry:
                handler_namespaces_rev[handler.ns] = handler.name
        # FIXME: should catch exceptions (?)
        expanded = self._exec({},(),{},
                              (PTmetal_handler.metal_expander,),
                              handler_namespaces_rev)
        self.write(expanded)
        if REQUEST is not None:
            REQUEST.set('text', self.read()) # May not equal 'text'!
            return self.pt_editForm(manage_tabs_message='Macros expanded')
Globals.InitializeClass(ZopePageTemplate)


manage_addPageTemplateForm = PageTemplateFile(
    'www/ptAdd', globals(), __name__='manage_addPageTemplateForm')

from urllib import quote

def manage_addPageTemplate(self, id, title=None, text=None,
                           REQUEST=None, submit=None):
    "Add a Page Template with optional file content."

    id = str(id)
    if REQUEST is None:
        self._setObject(id, ZopePageTemplate(id, text))
        ob = getattr(self, id)
        if title:
            ob.pt_setTitle(title)
        return ob
    else:
        file = REQUEST.form.get('file')
        headers = getattr(file, 'headers', None)
        if headers is None or not file.filename:
            zpt = ZopePageTemplate(id)
        else:
            zpt = ZopePageTemplate(id, file, headers.get('content_type'))
            
        self._setObject(id, zpt)

        try: u = self.DestinationURL()
        except: u = REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u,quote(id))
        REQUEST.RESPONSE.redirect(u+'/manage_main')
    return ''

def initialize(context):
    context.registerClass(
        ZopePageTemplate,
        permission='Add Page Templates',
        constructors=(manage_addPageTemplateForm,
                      manage_addPageTemplate),
        icon='www/altpt.png',
        )
    try:
        # hook AltPT up CMF's brain
        import FSAltPageTemplate
        FSAltPageTemplate.initialize(context)
    except ImportError:
        pass
